@@ -1,12 +1,13 @@
 """
 Main module containing functions.
 """
+from Bio import SeqIO
 from os import path
-from subprocess import Popen, PIPE
+from subprocess import Popen, PIPE, getoutput
 import pandas as pd
 
 
-def pipeline(table, taxonomy, metadata, output, column, n, export_qza=True):
+def filter_by_category(table, taxonomy, metadata, output, column, n, skip_qza=False):
     """
     Pipeline to convert Qiime 2 artifacts, add metadata and taxonomy, and filter most abundant ASVs.
     :param table: Feature table in QZA format from Qiime 2.
@@ -15,18 +16,18 @@ def pipeline(table, taxonomy, metadata, output, column, n, export_qza=True):
     :param output: Name of output directory.
     :param column: Name of metadata column to group by. Must be a column in the Manifest file.
     :param n: Get the n most abundant ASVs.
-    :param export_qza: Whether to export Qiime 2 artifacts.
+    :param skip_qza: Whether to skip converting Qiime 2 artifacts.
     :return:
     """
-    if export_qza:
+    if skip_qza:
+        biom_table = table
+        taxonomy_table = taxonomy
+    else:
         print("Converting Qiime 2 artifacts [...]")
         biom_table_dir = export_qiime_artifact(table)
         biom_table = path.join(biom_table_dir, "feature-table.biom")
         taxonomy_table_dir = export_qiime_artifact(taxonomy)
         taxonomy_table = path.join(taxonomy_table_dir, "taxonomy.tsv")
-    else:
-        biom_table = table
-        taxonomy_table = taxonomy
 
     print("Converting BIOM table [...]")
     feature_table = convert_biom_table(biom_table, _print=False)
@@ -62,6 +63,10 @@ def export_qiime_artifact(qza_file, _print=False):
     :param _print: pass to run_cmd
     :return: Output directory.
     """
+    qiime_bin = getoutput("qiime")
+    assert (
+        qiime_bin
+    ), "Could not detect Qiime 2 on your system. Make sure it is on $PATH"
     assert path.isfile(qza_file), f"{qza_file} does not exist!"
     qza_file, qza_file_abs = path.basename(qza_file), path.abspath(qza_file)
     out_dir = path.join(path.dirname(qza_file_abs), path.splitext(qza_file)[0])
@@ -150,6 +155,8 @@ def n_largest_by_category(
     cat=None, counts=None, metadata=None, n=0, category=None, run_merge=()
 ):
     """
+    Get the n most abundant ASVs by category.
+
     :param cat: Output of merge_metadata_and_taxonomy
     :param counts: Output of merge_metadata_and_taxonomy
     :param metadata: Output of merge_metadata_and_taxonomy
@@ -175,6 +182,7 @@ def n_largest_by_category(
     # Create output dictionary
     nlargest = dict()
 
+    # sample_as_categories or not
     cat = cat.T
     if category is None:
         category_values = samples
@@ -187,6 +195,7 @@ def n_largest_by_category(
         category_values = {i for i in cat[category].value_counts().index if i == i}
         samples_as_categories = False
 
+    # parsing categories
     for value in category_values:
         if samples_as_categories:
             grouped_df = cat.loc[value][asvs].astype("float")
@@ -195,9 +204,11 @@ def n_largest_by_category(
             grouped_df = cat[cat[category] == value][asvs].astype("float")
             grouped_counts = grouped_df.sum().nlargest(n)
             str_samples = " ".join(i for i in grouped_df.index if i in samples)
-            grouped_df["samples"] = str_samples
+            grouped_df["samples"] = str_samples  # add a column with sample names
 
-        grouped_counts = grouped_counts[grouped_counts > 0]
+        grouped_counts = grouped_counts[
+            grouped_counts > 0
+        ]  # sometimes there will be less than n ASVs in the sample
         grouped_asvs = list(grouped_counts.index)
         grouped_df = cat[grouped_asvs].copy().T
         grouped_df = grouped_df[tax_columns].copy()
@@ -208,3 +219,103 @@ def n_largest_by_category(
     cat_df = pd.concat((v for k, v in nlargest.items()))
 
     return cat_df
+
+
+def extract_asvs_from_fasta(seqids, fasta_file):
+    """
+    Subsets sequences from a fasta file and writes them to a new fasta file.
+    :param seqids: List of Seq IDs to extract
+    :param fasta_file: Path to FASTA file
+    :return:
+    """
+    assert path.isfile(fasta_file), "FASTA file does not exist!"
+    print("Extracting ASVs from {}".format(fasta_file))
+    records, subset = SeqIO.parse(fasta_file, "fasta"), []
+    seqids = set(seqids)  # Drop duplicates
+    no_subset = len(seqids)  # Number of sequences to subset
+    for i in records:
+        if i.id in seqids:
+            subset.append(i)
+
+    output_file, _ = path.splitext(fasta_file)
+    output_file += f"_subset_{no_subset}.fasta"
+    with open(output_file, "w") as f:
+        SeqIO.write(subset, f, "fasta")
+
+    if path.getsize(output_file) > 0:
+        print(f"Wrote {no_subset} sequences to {output_file}")
+
+    return output_file
+
+
+def blastn(query, db, params, _print=True):
+    """
+    Run blastn on a FASTA file.
+    :param query: Path to a fasta file.
+    :param db: Path to blast database.
+    :param params: string of params to pass to blastn
+    :param _print: Whether to print blast is running.
+    :return:
+    """
+    blastn_bin = getoutput("which blastn")
+    assert (
+        "command not found" not in blastn_bin
+    ), "Could not find blastn. Make sure blastn is on $PATH."
+    assert path.isfile(db), "Could not find db {}. Make sure db exists".format(db)
+    output_file, _ = path.splitext(query)
+    output_file += f"_blast_out.tsv"
+    outfmt = "'6 qseqid sseqid stitle pident length mismatch gapopen qstart qend sstart send evalue bitscore'"
+    cmd = (
+        blastn_bin
+        + f"-query {0} -db {1} -out {2} -outfmt {3} ".format(
+            query, db, output_file, outfmt
+        )
+        + params.strip()
+    )
+    if _print:
+        base_ = path.basename(query)
+        print("Running BLASTn for {}".format(base_))
+        print("$ {}".format(" ".join(path.basename(i) for i in cmd)))
+    run_cmd(cmd, output_file, _print=True)
+
+    return output_file
+
+
+def extract_asvs_and_blast(
+    asv_table, db, sequences, params="", skip_convert_sequences=False
+):
+    """
+    Runs BLASTn for the ASV table generate by filter_by_category.
+    :param asv_table: Output tsv of filter_by_category.
+    :param db: Path to BLAST db.
+    :param sequences: Sequences .qza or .fasta file.
+    :param params: parameters to pass to BLASTn
+    :param skip_convert_sequences: Skip converting sequences from .qza to .fasta.
+    :return:
+    """
+    asv_table = pd.read_csv(asv_table, sep="\t")
+    seqids = set(asv_table.iloc[:, 0])
+
+    if not skip_convert_sequences:
+        fasta_file = sequences
+    else:
+        fasta_file = path.join(export_qiime_artifact(sequences), "dna-sequences.fasta")
+
+    query = extract_asvs_from_fasta(seqids, fasta_file)
+    blast_out = blastn(query, db, params=params)
+    blast_out = add_header_to_blast_out(blast_out)
+
+    return blast_out
+
+
+def add_header_to_blast_out(blast_out):
+    """
+    Adds header row to blast_out
+    :param blast_out: Tab delimited blast_out file.
+    :return:
+    """
+    df = pd.read_csv(blast_out, sep="\t")
+    df.columns = "qseqid sseqid stitle pident length mismatch gapopen qstart qend sstart send evalue bitscore".split()
+    df.to_csv(blast_out, sep="\t", index=False)
+
+    return blast_out
